.js-plotly-plot .plotly .modebar {
  padding-top: 5%;
  margin-right: 3.5%;
}

body {
  background-color: #f2f2f2;
  margin-top: 1%;
  margin-left: 5%;
  margin-right: 5%;
  margin-bottom: 1%
}

.two.columns {
  width: 16.25%;
}

.column,
.columns {
  margin-left: 0.5%;
}

.pretty_container {
  border-radius: 5px;
  background-color: #f9f9f9;
  margin: 10px;
  padding: 15px;
  position: relative;
  box-shadow: 2px 2px 2px lightgrey;
}

.bare_container {
  margin: 0 0 0 0;
  padding: 0 0 0 0;
}

.dcc_control {
  margin: 0;
  padding: 5px;
  width: calc(100%-40px);
}

.control_label {
  margin: 0;
  padding: 10px;
  padding-bottom: 0px;
  margin-bottom: 0px;
  width: calc(100%-40px);
}

.rc-slider {
  margin-left: 0px;
  padding-left: 0px;
}

.flex-display {
  display: flex;
}

.container-display {
  display: flex;
}

#individual_graph,
#aggregate_graph {
  width: calc(100% - 30px);
  position: absolute;
}

#count_graph {
  position: absolute;
  height: calc(100% - 30px);
  width: calc(100% - 30px);
}

#countGraphContainer {
  flex: 5;
  position: relative;
}

#header {
  align-items: center;
}

#learn-more-button {
  text-align: center;
  height: 100%;
  padding: 0 20px;
  text-transform: none;
  font-size: 15px;
  float: right;
  margin-right: 10px;
  margin-top: 30px;
}
#title {
  text-align: center;
}

.mini_container {
  border-radius: 5px;
  background-color: #f9f9f9;
  margin: 10px;
  padding: 15px;
  position: relative;
  box-shadow: 2px 2px 2px lightgrey;
}

#right-column {
  display: flex;
  flex-direction: column;
}

#wells {
  flex: 1;
}

#gas {
  flex: 1;
}

#aggregate_data {
  align-items: center;
}

#oil {
  flex: 1;
}

#water {
  flex: 1;
}

#tripleContainer {
  display: flex;
  flex: 3;
}

#mainContainer {
  display: flex;
  flex-direction: column;
}

#pie_graph > div > div > svg:nth-child(3) > g.infolayer > g.legend {
  pointer-events: all;
  transform: translate(30px, 349px);
}
